{% include header.html %}

<div class="navbar white">
  <div class="name"><a href="/">containerd<span class="org">By Docker</span></a></div>
  <div class="actions">
      {% include links.html %}
  </div>
</div>
<div class="content">
  <div class="wrapper-details">
    {{ content }}
  </div>
</div>

{% include footer.html %}
