{% include header.html %}

<div class="hero">
  <div class="navbar">
    <div class="name"><a href="/">containerd</a></div>
    <div class="actions">
      {% include links.html %}
    </div>
  </div>
  <p class="definition"><strong>containerd</strong> <span>| kənˈtānər D |</span><br> - an industry-standard container runtime<br>with an emphasis on <i>simplicity</i>,<br><i>robustness</i> and <i>portability</i>.</p>
</div>
<div class="content">
  <div class="wrapper-details">
      {{ content }}
  </div>
</div>

{% include footer.html %}
