<!DOCTYPE html>
<html lang="en">
  <head>
    <meta http-equiv="Content-Type" content="text/html; charset=UTF-8"/>
    <meta name="viewport" content="width=device-width, initial-scale=1">
    <title>{{ page.title }}</title>
    <link rel="icon" type="image/png" href="/images/favicon-16x16.png" sizes="16x16">
    <link href="/style/css/flexboxgrid.css" type="text/css" rel="stylesheet" media="screen,projection"/>
    <link href="/style/css/main.css" type="text/css" rel="stylesheet" media="screen,projection"/>
    <link href="/style/css/tomorrow-night.css" type="text/css" rel="stylesheet" media="screen,projection"/><link href="https://fonts.googleapis.com/css?family=Lato:300,400|Raleway:400,700i,800" rel="stylesheet">
  </head>
  <body>
    <div class="container">
