@font-regular:          "Helvetica Neue", Segoe UI, Arial, "Lucida Grande", sans-serif;

@brand-primary:         #00DA9C;

@gray-darkest:          #254451;
@gray-dark:             #364449;

@blue-dark:             #0077A8;

@color-background:      #E7F4F9;
@color-gradient-start:  #0078C5;
@color-gradient-end:    #004775;